@@ -2,39 +2,48 @@
 
 from __future__ import annotations
 
+import functools
 import logging
 import os
 import pathlib
 import shutil
+import sys
 import warnings
 from abc import ABC, abstractmethod
-from typing import TYPE_CHECKING, Any, Sequence
+from collections.abc import Sequence
+from typing import TYPE_CHECKING, Any
 
-log = logging.getLogger(__name__)
+if sys.version_info >= (3, 12):
+    from typing import override
+else:
+    from typing_extensions import override
 
-from overrides import overrides
+logger = logging.getLogger(__name__)
+
+from pycontrails.utils import dependencies
 
 # optional imports
 if TYPE_CHECKING:
-    import google
+    import google.cloud.storage
 
 
+@functools.cache
 def _get_user_cache_dir() -> str:
     try:
         import platformdirs
     except ModuleNotFoundError as e:
-        raise ModuleNotFoundError(
-            "Using the pycontrails CacheStore requires the 'platformdirs' package. "
-            "This can be installed with 'pip install pycontrails[ecmwf]' or "
-            "'pip install platformdirs'."
-        ) from e
+        dependencies.raise_module_not_found_error(
+            name="cache module",
+            package_name="platformdirs",
+            module_not_found_error=e,
+        )
     return platformdirs.user_cache_dir("pycontrails")
 
 
 class CacheStore(ABC):
     """Abstract cache storage class for storing staged and intermediate data."""
 
-    __slots__ = ("cache_dir", "allow_clear")
+    __slots__ = ("allow_clear", "cache_dir")
     cache_dir: str
     allow_clear: bool
 
@@ -51,17 +60,11 @@ class CacheStore(ABC):
         Examples
         --------
         >>> from pycontrails import DiskCacheStore
-        >>> disk_cache = DiskCacheStore(cache_dir="cache", allow_clear=True)
-        >>> disk_cache.size
-        0.0...
-
-        >>> disk_cache.clear()  # cleanup
-
-        >>> from pycontrails import GCPCacheStore
-        >>> gcp_cache = GCPCacheStore(bucket="contrails-301217-unit-test", cache_dir="cache")
-        >>> gcp_cache.size
+        >>> cache = DiskCacheStore(cache_dir="cache", allow_clear=True)
+        >>> cache.size
         0.0...
 
+        >>> cache.clear()  # cleanup
         """
 
     @abstractmethod
@@ -97,16 +100,11 @@ class CacheStore(ABC):
         Examples
         --------
         >>> from pycontrails import DiskCacheStore
-        >>> disk_cache = DiskCacheStore(cache_dir="cache", allow_clear=True)
-        >>> disk_cache.path("file.nc")
+        >>> cache = DiskCacheStore(cache_dir="cache", allow_clear=True)
+        >>> cache.path("file.nc")
         'cache/file.nc'
 
-        >>> disk_cache.clear()  # cleanup
-
-        >>> from pycontrails import GCPCacheStore
-        >>> gcp_cache = GCPCacheStore(bucket="contrails-301217-unit-test", cache_dir="cache")
-        >>> gcp_cache.path("file.nc")
-        'cache/file.nc'
+        >>> cache.clear()  # cleanup
         """
 
     @abstractmethod
@@ -126,13 +124,8 @@ class CacheStore(ABC):
         Examples
         --------
         >>> from pycontrails import DiskCacheStore
-        >>> disk_cache = DiskCacheStore(cache_dir="cache", allow_clear=True)
-        >>> disk_cache.exists("file.nc")
-        False
-
-        >>> from pycontrails import GCPCacheStore
-        >>> gcp_cache = GCPCacheStore(bucket="contrails-301217-unit-test", cache_dir="cache")
-        >>> gcp_cache.exists("file.nc")
+        >>> cache = DiskCacheStore(cache_dir="cache", allow_clear=True)
+        >>> cache.exists("file.nc")
         False
         """
 
@@ -157,7 +150,7 @@ class CacheStore(ABC):
         """
 
         # TODO: run in parallel?
-        return [self.put(d, cp) for d, cp in zip(data_path, cache_path)]
+        return [self.put(d, cp) for d, cp in zip(data_path, cache_path, strict=True)]
 
     # In the three methods below, child classes have a complete docstring.
 
@@ -175,12 +168,12 @@ class DiskCacheStore(CacheStore):
 
     Parameters
     ----------
-    allow_clear : bool, optional
-        Allow this cache to be cleared using :meth:`clear()`. Defaults to False.
-    cache_dir : str | pathlib.Path, optional
+    cache_dir : str | pathlib.Path | None, optional
         Root cache directory.
         By default, looks first for ``PYCONTRAILS_CACHE_DIR`` environment variable,
         then uses the OS specific :func:`platformdirs.user_cache_dir` function.
+    allow_clear : bool, optional
+        Allow this cache to be cleared using :meth:`clear()`. Defaults to False.
 
     Examples
     --------
@@ -198,11 +191,8 @@ class DiskCacheStore(CacheStore):
         allow_clear: bool = False,
     ):
         if cache_dir is None:
-            # Use a try / except to avoid unnecessary import of platformdirs
-            try:
-                cache_dir = os.environ["PYCONTRAILS_CACHE_DIR"]
-            except KeyError:
-                cache_dir = _get_user_cache_dir()
+            # Avoid unnecessary import of platformdirs (called in _get_user_cache_dir)
+            cache_dir = os.getenv("PYCONTRAILS_CACHE_DIR") or _get_user_cache_dir()
 
         # make sure local cache directory exists
         pathlib.Path(cache_dir).mkdir(parents=True, exist_ok=True)
@@ -217,37 +207,35 @@ class DiskCacheStore(CacheStore):
         return f"DiskCacheStore: {self.cache_dir}"
 
     @property
-    @overrides
+    @override
     def size(self) -> float:
         disk_path = pathlib.Path(self.cache_dir)
-        size = sum(f.stat().st_size for f in disk_path.glob("**/*") if f.is_file())
-        log.debug(f"Disk cache size {size} bytes")
+        size = sum(f.stat().st_size for f in disk_path.rglob("*") if f.is_file())
+        logger.debug("Disk cache size %s bytes", size)
         return size / 1e6
 
-    @overrides
+    @override
     def listdir(self, path: str = "") -> list[str]:
         path = self.path(path)
         iter_ = pathlib.Path(path).iterdir()
-        out = [str(f.relative_to(path)) for f in iter_]
-        out.sort()
-        return out
+        return sorted(str(f.relative_to(path)) for f in iter_)
 
-    @overrides
+    @override
     def path(self, cache_path: str) -> str:
-        if str(cache_path).startswith(str(self.cache_dir)):
+        if cache_path.startswith(self.cache_dir):
             disk_path = pathlib.Path(cache_path)
         else:
-            disk_path = pathlib.Path(self.cache_dir) / pathlib.Path(cache_path)
+            disk_path = pathlib.Path(self.cache_dir) / cache_path
 
         # make sure full path to parents exist
         disk_path.parent.mkdir(parents=True, exist_ok=True)
 
         return str(disk_path)
 
-    @overrides
+    @override
     def exists(self, cache_path: str) -> bool:
         disk_path = pathlib.Path(self.path(cache_path))
-        return disk_path.is_dir() or disk_path.is_file()
+        return disk_path.exists()
 
     def put(self, data_path: str | pathlib.Path, cache_path: str | None = None) -> str:
         """Save data to the local cache store.
@@ -278,6 +266,8 @@ class DiskCacheStore(CacheStore):
         >>> # put a file directly
         >>> disk_cache.put("README.md", "test/file.md")
         'test/file.md'
+
+        >>> disk_cache.clear()  # cleanup
         """
 
         if not pathlib.Path(data_path).is_file():
@@ -289,15 +279,22 @@ class DiskCacheStore(CacheStore):
         disk_path = self.path(str(cache_path))
 
         # copy to disk cache
-        log.debug(f"Disk cache put {data_path} to {disk_path} in disk cache")
-        shutil.copy(data_path, disk_path)
+        logger.debug("Disk cache put %s to %s in disk cache", data_path, cache_path)
+        try:
+            shutil.copyfile(data_path, disk_path)
+        except PermissionError:
+            logger.warning(
+                "Permission error copying %s to %s. The destination file may already be open.",
+                data_path,
+                disk_path,
+            )
 
         return cache_path
 
     def get(self, cache_path: str) -> str:
         """Get data path from the local cache store.
 
-        Alias for :meth:`path()`
+        Alias for :meth:`path`
 
         Parameters
         ----------
@@ -317,6 +314,8 @@ class DiskCacheStore(CacheStore):
         >>> # returns a path
         >>> disk_cache.get("test/file.md")
         'cache/test/file.md'
+
+        >>> disk_cache.clear()  # cleanup
         """
         return self.path(cache_path)
 
@@ -361,20 +360,18 @@ class DiskCacheStore(CacheStore):
         disk_path = pathlib.Path(self.path(cache_path))
 
         if disk_path.is_file():
-            log.debug("Remove file at path %s", disk_path)
+            logger.debug("Remove file at path %s", disk_path)
             disk_path.unlink()
+            return
 
         # Assume anything else is a directory
-        elif disk_path.exists():
+        if disk_path.exists():
             # rm directory recursively
-            log.debug("Remove directory at path %s", disk_path)
+            logger.debug("Remove directory at path %s", disk_path)
             shutil.rmtree(disk_path, ignore_errors=True)
+            return
 
-        else:
-            warnings.warn(f"No cache path found at {disk_path}")
-
-        # make sure local cache directory exists
-        # pathlib.Path(self.cache_dir).mkdir(parents=True, exist_ok=True)
+        warnings.warn(f"No cache path found at {disk_path}")
 
 
 class GCPCacheStore(CacheStore):
@@ -422,7 +419,10 @@ class GCPCacheStore(CacheStore):
     Examples
     --------
     >>> from pycontrails import GCPCacheStore
-    >>> cache = GCPCacheStore(bucket="contrails-301217-unit-test", cache_dir="cache")
+    >>> cache = GCPCacheStore(
+    ...     bucket="contrails-301217-unit-test",
+    ...     cache_dir="cache",
+    ... )
     >>> cache.cache_dir
     'cache/'
     >>> cache.bucket
@@ -430,15 +430,15 @@ class GCPCacheStore(CacheStore):
     """
 
     __slots__ = (
-        "project",
-        "bucket",
-        "read_only",
-        "timeout",
-        "show_progress",
-        "chunk_size",
-        "_disk_cache",
-        "_client",
         "_bucket",
+        "_client",
+        "_disk_cache",
+        "bucket",
+        "chunk_size",
+        "project",
+        "read_only",
+        "show_progress",
+        "timeout",
     )
     project: str | None
     bucket: str
@@ -447,8 +447,8 @@ class GCPCacheStore(CacheStore):
     show_progress: bool
     chunk_size: int
     _disk_cache: DiskCacheStore
-    _client: "google.cloud.storage.Client"
-    _bucket: "google.cloud.storage.Bucket"
+    _client: google.cloud.storage.Client
+    _bucket: google.cloud.storage.Bucket
 
     def __init__(
         self,
@@ -465,10 +465,11 @@ class GCPCacheStore(CacheStore):
         try:
             from google.cloud import storage
         except ModuleNotFoundError as e:
-            raise ModuleNotFoundError(
-                "GCPCache requires the `google-cloud-storage` module, which can be installed "
-                "using `pip install pycontrails[gcp]`"
-            ) from e
+            dependencies.raise_module_not_found_error(
+                name="GCPCacheStore class",
+                package_name="google-cloud-storage",
+                module_not_found_error=e,
+            )
 
         if "https://" in cache_dir:
             raise ValueError(
@@ -508,7 +509,7 @@ class GCPCacheStore(CacheStore):
             )
 
         # append a "/" for GCP objects
-        if cache_dir != "" and not cache_dir.endswith("/"):
+        if cache_dir and not cache_dir.endswith("/"):
             cache_dir = f"{cache_dir}/"
 
         # set up gcp client
@@ -547,7 +548,7 @@ class GCPCacheStore(CacheStore):
         return f"GCPCacheStore: {self.bucket}/{self.cache_dir}"
 
     @property
-    def client(self) -> "google.cloud.storage.Client":
+    def client(self) -> google.cloud.storage.Client:
         """Handle to Google Cloud Storage client.
 
         Returns
@@ -558,15 +559,15 @@ class GCPCacheStore(CacheStore):
         return self._client
 
     @property
-    @overrides
+    @override
     def size(self) -> float:
         # get list of blobs below this path
         blobs = self._bucket.list_blobs(prefix=self.cache_dir)
         size = sum(b.size for b in blobs)
-        log.debug(f"GCP cache size {size} bytes")
+        logger.debug("GCP cache size %s bytes", size)
         return size / 1e6
 
-    @overrides
+    @override
     def listdir(self, path: str = "") -> list[str]:
         # I don't necessarily think we want to implement this .... it might be
         # very slow if the bucket is large. BUT, it won't be slower than the size
@@ -579,7 +580,7 @@ class GCPCacheStore(CacheStore):
             "list files in the local disk cache."
         )
 
-    @overrides
+    @override
     def path(self, cache_path: str) -> str:
         if cache_path.startswith(self.cache_dir):
             return cache_path
@@ -601,18 +602,21 @@ class GCPCacheStore(CacheStore):
         Examples
         --------
         >>> from pycontrails import GCPCacheStore
-        >>> cache = GCPCacheStore(bucket="contrails-301217-unit-test", cache_dir="cache")
+        >>> cache = GCPCacheStore(
+        ...     bucket="contrails-301217-unit-test",
+        ...     cache_dir="cache",
+        ... )
         >>> cache.path("file.nc")
         'cache/file.nc'
         """
         bucket_path = self.path(cache_path)
         return f"gs://{self.bucket}/{bucket_path}"
 
-    @overrides
+    @override
     def exists(self, cache_path: str) -> bool:
         # see if file is in the mirror disk cache
         if self._disk_cache.exists(cache_path):
-            return self._disk_cache.exists(cache_path)
+            return True
 
         bucket_path = self.path(cache_path)
         blob = self._bucket.blob(bucket_path)
@@ -663,7 +667,7 @@ class GCPCacheStore(CacheStore):
 
         # read only
         if self.read_only:
-            log.debug(
+            logger.debug(
                 f"GCP Cache Store is read only. File put in local DiskCacheStore path: {cache_path}"
             )
             raise RuntimeError(
@@ -676,7 +680,7 @@ class GCPCacheStore(CacheStore):
         disk_path = self._disk_cache.path(cache_path)
         blob = self._bucket.blob(bucket_path)
 
-        log.debug(f"GCP Cache put {disk_path} to {bucket_path}")
+        logger.debug("GCP Cache put %s to %s", disk_path, bucket_path)
 
         if self.show_progress:  # upload with pbar
             _upload_with_progress(blob, disk_path, self.timeout, chunk_size=self.chunk_size)
@@ -739,7 +743,7 @@ class GCPCacheStore(CacheStore):
         if not blob.exists():
             raise ValueError(f"No object exists in cache at path {bucket_path}")
 
-        log.debug(f"GCP Cache GET from {bucket_path}")
+        logger.debug("GCP Cache GET from %s", bucket_path)
 
         if self.show_progress:
             _download_with_progress(
@@ -765,8 +769,11 @@ class GCPCacheStore(CacheStore):
         Examples
         --------
         >>> from pycontrails import GCPCacheStore
-        >>> gcp_cache = GCPCacheStore(bucket="contrails-301217-unit-test", cache_dir="cache")
-        >>> gcp_cache.clear_disk()
+        >>> cache = GCPCacheStore(
+        ...     bucket="contrails-301217-unit-test",
+        ...     cache_dir="cache",
+        ... )
+        >>> cache.clear_disk()
         """
         self._disk_cache.clear(cache_path)
 
@@ -794,7 +801,7 @@ class GCPCacheStore(CacheStore):
 
         # get full path to clear
         bucket_path = self.path(cache_path)
-        log.debug(f"Clearing GCP cache at path {str(bucket_path)}")
+        logger.debug("Clearing GCP cache at path %s", bucket_path)
 
         # clear disk mirror
         self.clear_disk()
@@ -827,11 +834,12 @@ def _upload_with_progress(blob: Any, disk_path: str, timeout: int, chunk_size: i
     try:
         from tqdm.auto import tqdm
     except ModuleNotFoundError as e:
-        raise ModuleNotFoundError(
-            "Method `put` requires the `tqdm` module, which can be installed using "
-            "`pip install pycontrails[gcp]`. "
-            "Alternatively, set instance attribute `show_progress=False`."
-        ) from e
+        dependencies.raise_module_not_found_error(
+            name="_upload_with_progress function",
+            package_name="tqdm",
+            module_not_found_error=e,
+            pycontrails_optional_package="gcp",
+        )
 
     # minimal possible chunk_size to allow nice progress bar
     blob.chunk_size = chunk_size
@@ -849,21 +857,33 @@ def _download_with_progress(
 
     try:
         from google.resumable_media.requests import ChunkedDownload
+    except ModuleNotFoundError as e:
+        dependencies.raise_module_not_found_error(
+            name="_download_with_progress function",
+            package_name="google-cloud-storage",
+            module_not_found_error=e,
+            pycontrails_optional_package="gcp",
+        )
+
+    try:
         from tqdm.auto import tqdm
     except ModuleNotFoundError as e:
-        raise ModuleNotFoundError(
-            "Method `get` requires the `tqdm` and `google-cloud-storage` modules, "
-            "which can be installed using `pip install pycontrails[gcp]`. "
-            "Alternatively, set instance attribute `show_progress=False`."
-        ) from e
+        dependencies.raise_module_not_found_error(
+            name="_download_with_progress function",
+            package_name="tqdm",
+            module_not_found_error=e,
+            pycontrails_optional_package="gcp",
+        )
 
     blob = gcp_cache._bucket.get_blob(gcp_path)
     url = blob._get_download_url(gcp_cache._client)
     description = f"Download {gcp_path}"
 
-    with open(disk_path, "wb") as local_file:
-        with tqdm.wrapattr(local_file, "write", total=blob.size, desc=description) as file_obj:
-            download = ChunkedDownload(url, chunk_size, file_obj)
-            transport = gcp_cache.client._http
-            while not download.finished:
-                download.consume_next_chunk(transport, timeout=gcp_cache.timeout)
+    with (
+        open(disk_path, "wb") as local_file,
+        tqdm.wrapattr(local_file, "write", total=blob.size, desc=description) as file_obj,
+    ):
+        download = ChunkedDownload(url, chunk_size, file_obj)
+        transport = gcp_cache.client._http
+        while not download.finished:
+            download.consume_next_chunk(transport, timeout=gcp_cache.timeout)
