@@ -28,7 +28,7 @@ def test_sl_path_to_met(met_ecmwf_sl_path: str) -> None:
     assert list(ds.data_vars) == ["sp"]
 
     mds = MetDataset(ds)
-    assert mds.dim_order == ["longitude", "latitude", "level", "time"]
+    assert mds.dim_order == ("longitude", "latitude", "level", "time")
 
     # longitude and latitude correctly translated
     assert np.all(mds.data["longitude"].values >= -180)
@@ -73,7 +73,7 @@ def test_pl_path_to_met(met_ecmwf_pl_path: str) -> None:
     assert list(ds.data_vars) == ["t", "r", "q", "ciwc"]
 
     mds = MetDataset(ds)
-    assert mds.dim_order == ["longitude", "latitude", "level", "time"]
+    assert mds.dim_order == ("longitude", "latitude", "level", "time")
 
     # longitude and latitude correctly translated
     assert np.all(mds.data["longitude"].values < 180)
@@ -89,10 +89,10 @@ def test_pl_path_to_met(met_ecmwf_pl_path: str) -> None:
         shapes = shapes + mds.data[coord].shape
     assert shapes == mds.data["t"].values.shape
 
-    with pytest.raises(ValueError, match="Set 'copy=True' when using 'wrap_longitude=True'."):
+    with pytest.raises(ValueError, match=r"Set 'copy=True' when using 'wrap_longitude=True'."):
         MetDataset(ds, wrap_longitude=True, copy=False)
 
-    with pytest.raises(ValueError, match="Coordinate 'latitude' not sorted."):
+    with pytest.raises(ValueError, match=r"Coordinate 'latitude' not sorted."):
         MetDataset(ds, copy=False)
 
     assert not mds.is_zarr
@@ -122,13 +122,14 @@ def test_metdataarray_constructor(
     assert mda2.is_wrapped
     assert mda2.shape[0] == 17
 
-    with pytest.raises(ValueError, match="Set 'copy=True' when using 'wrap_longitude=True'."):
+    with pytest.raises(ValueError, match=r"Set 'copy=True' when using 'wrap_longitude=True'."):
         MetDataArray(mda.data, wrap_longitude=True, copy=False)
 
     # Cannot instantiate MetDataArray without time or level coord
     # This hacked DataArray only contains latitude and longitude
     insufficient_da = da["latitude"] + da["longitude"]
-    with pytest.raises(ValueError, match="Meteorology data must contain dimension 'level'"):
+    match = r"Meteorology data must contain dimension\(s\): \['level', 'time'\]."
+    with pytest.raises(ValueError, match=match):
         MetDataArray(insufficient_da)
 
 
@@ -158,7 +159,7 @@ def test_pl_path_to_met_dataset(met_ecmwf_pl_path: str, wrap_longitude: bool):
 
 
 @pytest.mark.parametrize(
-    "lon,lat,match",
+    ("lon", "lat", "match"),
     [
         (np.linspace(-200, 200, 10), np.linspace(-100, 100, 10), "Latitude values must"),
         (np.linspace(-400, 200, 10), np.linspace(-80, 80, 10), "Shift to WGS84"),
@@ -215,26 +216,80 @@ def test_wrap_longitude(zero_like_da: xr.DataArray, met_ecmwf_pl_path: str) -> N
 def test_shift_longitude(zero_like_da: xr.DataArray) -> None:
     da = zero_like_da.copy()
 
+    # assign random values to all coords
+    rng = np.random.default_rng()
+    da.loc[:] = rng.random(size=da.shape)
+
+    # shift coordinates to [0, 360) manually
     lons = da["longitude"].values
-    lons[lons < 0] = lons[lons < 0] + 360
+    lons = np.where(lons < 0.0, lons + 360.0, lons)
 
     da = da.assign_coords(longitude=lons)
     assert np.all(da["longitude"].values >= 0)
 
     da2 = shift_longitude(da)
 
+    # longitude sorted
+    assert np.all(np.diff(da2["longitude"].values) > 0)
+
     # longitude correctly translated
-    assert np.all(da2["longitude"].values <= 180)
+    assert np.all(da2["longitude"].values < 180)
     assert np.all(da2["longitude"].values >= -180)
 
+    # test a few values
+    assert (
+        da.sel({"longitude": 181})[0, 0, 0].values == da2.sel({"longitude": -179})[0, 0, 0].values
+    )
+    assert (
+        da.sel({"longitude": 190})[0, 0, 0].values == da2.sel({"longitude": -170})[0, 0, 0].values
+    )
+
+
+def test_shift_longitude_bound(zero_like_da: xr.DataArray) -> None:
+    da = zero_like_da.copy()
+
+    # assign random values to all coords
+    rng = np.random.default_rng()
+    da.loc[:] = rng.random(size=da.shape)
+
+    da2 = shift_longitude(da, bound=0)
+
     # longitude sorted
     assert np.all(np.diff(da2["longitude"].values) > 0)
 
+    # longitude correctly translated
+    assert np.all(da2["longitude"].values < 360.0)
+    assert np.all(da2["longitude"].values >= 0.0)
 
-@pytest.fixture
+    # test a few values
+    assert (
+        da.sel({"longitude": -179})[0, 0, 0].values == da2.sel({"longitude": 181})[0, 0, 0].values
+    )
+    assert (
+        da.sel({"longitude": -170})[0, 0, 0].values == da2.sel({"longitude": 190})[0, 0, 0].values
+    )
+
+    da2 = shift_longitude(da, bound=-10)
+
+    # longitude sorted
+    assert np.all(np.diff(da2["longitude"].values) > 0)
+
+    # longitude correctly translated
+    assert np.all(da2["longitude"].values < 350.0)
+    assert np.all(da2["longitude"].values >= -10.0)
+
+    # test a few values
+    assert da.sel({"longitude": -10})[0, 0, 0].values == da2.sel({"longitude": -10})[0, 0, 0].values
+    assert da.sel({"longitude": -20})[0, 0, 0].values == da2.sel({"longitude": 340})[0, 0, 0].values
+    assert (
+        da.sel({"longitude": -179})[0, 0, 0].values == da2.sel({"longitude": 181})[0, 0, 0].values
+    )
+
+
+@pytest.fixture()
 def zero_like_da() -> xr.DataArray:
     return xr.DataArray(
-        data=0,
+        data=0.0,
         dims=["longitude", "latitude", "level", "time"],
         coords={
             "longitude": np.arange(-180, 180, 1.0),
@@ -267,21 +322,33 @@ def sparse_binary(zero_like_da: xr.DataArray, request: Any) -> tuple[MetDataArra
     lat = request.param["lat"]
     level = request.param["level"]
     point = (da["latitude"] == lat) & (da["longitude"] == lon) & (da["level"] == level)
-    da = xr.where(point, 1, da)
+    da = xr.where(point, 1.0, da)
     return MetDataArray(da), request.param
 
 
-@pytest.fixture
+@pytest.fixture()
 def island_binary(zero_like_da: xr.DataArray) -> MetDataArray:
     # create 9 x 9 island
 
     da = zero_like_da
     island = (np.abs(da["latitude"]) < 5) & (np.abs(da["longitude"]) < 5)
-    da = xr.where(island, 1, da)
+    da = xr.where(island, 1.0, da)
     return MetDataArray(da)
 
 
-@pytest.fixture
+@pytest.fixture()
+def island_binary_nan(zero_like_da: xr.DataArray) -> MetDataArray:
+    # create 9 x 9 island and 5 x 5 region of nan
+
+    da = zero_like_da
+    island = (np.abs(da["latitude"]) < 5) & (np.abs(da["longitude"]) < 5)
+    da = xr.where(island, 1.0, da)
+    nan_island = (da["latitude"] >= 10) & (da["latitude"] < 15) & (np.abs(da["longitude"]) < 3)
+    da = xr.where(nan_island, np.nan, da)
+    return MetDataArray(da)
+
+
+@pytest.fixture()
 def antimeridian_binary(zero_like_da: xr.DataArray) -> MetDataArray:
     # create 9 x 9 island spanning antimeridian
 
@@ -319,12 +386,12 @@ def _generate_waypoints(
     return _x, _y, _z, _t
 
 
-@pytest.fixture
+@pytest.fixture()
 def pl_path_to_doubled_time_mds(met_ecmwf_pl_path: str) -> MetDataset:
     return MetDataset(xr.open_dataset(met_ecmwf_pl_path))
 
 
-@pytest.fixture
+@pytest.fixture()
 def sl_path_to_tripled_time_mds(met_ecmwf_sl_path: str) -> MetDataset:
     """Triple the number of time variables in the Dataset sitting at `met_ecmwf_sl_path`."""
     ds1 = xr.open_dataset(met_ecmwf_sl_path)
@@ -363,7 +430,7 @@ def test_met_values_property(zero_like_da: xr.DataArray, met_ecmwf_pl_path: str)
 
 
 @pytest.mark.parametrize(
-    "values, dimension",
+    ("values", "dimension"),
     [
         ((-211, 53, 250, np.datetime64("2019-05-31T05:30")), 0),
         ((26, -112, 250, np.datetime64("2019-05-31T05:30")), 1),
@@ -452,19 +519,21 @@ def test_interpolate_antimeridian_binary(antimeridian_binary: MetDataArray) -> N
 def test_bad_interpolate(island_binary: MetDataArray) -> None:
     assert len(island_binary.interpolate(5, 5, 200, np.datetime64("2019-05-31T05:00:00"))) == 1
 
-    with pytest.raises(ValueError):
+    match = "One of the requested xi is out of bounds in dimension"
+
+    with pytest.raises(ValueError, match=match):
         island_binary.interpolate(
             200, 5, 200, np.datetime64("2019-05-31T05:00:00"), bounds_error=True
         )
-    with pytest.raises(ValueError):
+    with pytest.raises(ValueError, match=match):
         island_binary.interpolate(
             5, 100, 200, np.datetime64("2019-05-31T05:00:00"), bounds_error=True
         )
-    with pytest.raises(ValueError):
+    with pytest.raises(ValueError, match=match):
         island_binary.interpolate(
             5, 5, 600, np.datetime64("2019-05-31T05:00:00"), bounds_error=True
         )
-    with pytest.raises(ValueError):
+    with pytest.raises(ValueError, match=match):
         island_binary.interpolate(
             5, 5, 200, np.datetime64("2021-05-31T05:00:00"), bounds_error=True
         )
@@ -518,7 +587,7 @@ def test_edges_on_median_binary(median_binary: MetDataArray) -> None:
 
 
 def test_polygons_sparse_binary_specify_time_level(
-    sparse_binary: tuple[MetDataArray, dict]
+    sparse_binary: tuple[MetDataArray, dict],
 ) -> None:
     mda, _ = sparse_binary
     # must specify level and time
@@ -562,14 +631,21 @@ def test_polygon_sparse_binary(
     mda = MetDataArray(mda.data.isel(time=[time]))
     level = nonzero_coord["level"]
 
-    geojson = mda.to_polygon_feature(level=level, min_area=0.0, iso_value=0.5, epsilon=0.0)
+    geojson = mda.to_polygon_feature(
+        level=level,
+        min_area=0.0,
+        iso_value=0.5,
+        epsilon=0.0,
+        precision=6,
+    )
     assert geojson["type"] == "Feature"
     assert geojson["geometry"]["type"] == "MultiPolygon"
     coords = geojson["geometry"]["coordinates"]
     assert isinstance(coords, list)
 
     # multipolygons are arrays of Polygons
-    assert len(coords) == 1 and len(coords[0])
+    assert len(coords) == 1
+    assert len(coords[0])
     coords = coords[0][0]
 
     # polygon closes in on itself
@@ -590,7 +666,7 @@ def test_polygon_sparse_binary(
     assert json.dumps(geojson)
 
 
-@pytest.fixture
+@pytest.fixture()
 def island_slice(island_binary: MetDataArray) -> MetDataArray:
     # every (level, time) slice of island_binary is identical -- just grab one
     return MetDataArray(island_binary.data.isel(level=[0], time=[0]))
@@ -602,7 +678,8 @@ def test_polygon_island_binary(island_slice: MetDataArray) -> None:
     # contains a single polygon
     coords = geojson["geometry"]["coordinates"]
     assert isinstance(coords, list)
-    assert len(coords) == 1 and len(coords[0]) == 1
+    assert len(coords) == 1
+    assert len(coords[0]) == 1
     coords = coords[0][0]
 
     # polygon closes in on itself
@@ -618,7 +695,57 @@ def test_polygon_island_binary(island_slice: MetDataArray) -> None:
             assert component in [-4.5, -4, 4, 4.5]
 
 
-def test_nested_polygons(zero_like_da: xr.DataArray) -> None:
+@pytest.fixture()
+def island_slice_nan(island_binary_nan: MetDataArray) -> MetDataArray:
+    # every (level, time) slice of island_binary_nan is identical -- just grab one
+    return MetDataArray(island_binary_nan.data.isel(level=[0], time=[0]))
+
+
+def test_polygon_island_nan_handling(island_slice_nan: MetDataArray) -> None:
+    """Test for identical results with retained nan and fill value below threshold"""
+
+    # expect identical results with finite fill value below threshold...
+    geojson1 = island_slice_nan.to_polygon_feature(
+        iso_value=0.5, fill_value=np.nan, epsilon=0.01, precision=2
+    )
+    geojson2 = island_slice_nan.to_polygon_feature(
+        iso_value=0.5, fill_value=0.0, epsilon=0.01, precision=2
+    )
+    assert geojson1 == geojson2
+
+    # ... but not with finite fill value above threshold
+    geojson3 = island_slice_nan.to_polygon_feature(
+        iso_value=0.5, fill_value=1.0, epsilon=0.01, precision=2
+    )
+    assert geojson1 != geojson3
+
+
+@pytest.fixture()
+def island_slice_neg(island_binary_nan: MetDataArray) -> MetDataArray:
+    # negate values in slice of island_binary_nan
+    return MetDataArray(-island_binary_nan.data.isel(level=[0], time=[0]))
+
+
+def test_polygon_island_lower_upper_bound(
+    island_slice_nan: MetDataArray, island_slice_neg: MetDataArray
+) -> None:
+    """Test polygon generation with lower and upper bound."""
+
+    geojson1 = island_slice_nan.to_polygon_feature(
+        iso_value=0.5, lower_bound=True, epsilon=0.01, precision=2
+    )
+    geojson2 = island_slice_neg.to_polygon_feature(
+        iso_value=-0.5, lower_bound=True, epsilon=0.01, precision=2
+    )
+    geojson3 = island_slice_neg.to_polygon_feature(
+        iso_value=-0.5, lower_bound=False, epsilon=0.01, precision=2
+    )
+    assert geojson1 != geojson2
+    assert geojson1 == geojson3
+
+
+@pytest.mark.parametrize("interiors", [True, False])
+def test_nested_polygons(zero_like_da: xr.DataArray, interiors: bool) -> None:
     """This test shows that we *dont* currently support deeply nested polygons.
 
     To update once we support this feature.
@@ -638,12 +765,17 @@ def test_nested_polygons(zero_like_da: xr.DataArray) -> None:
     da = xr.where((island5), 1, da)
 
     mda = MetDataArray(da.isel(level=[0], time=[0]))
-    geojson = mda.to_polygon_feature(iso_value=0.5)
+    geojson = mda.to_polygon_feature(iso_value=0.5, interiors=interiors)
 
     # contains 1 Polygon with 1 nested polygon
     coords = geojson["geometry"]["coordinates"]
     assert isinstance(coords, list)
-    assert len(coords) == 1 and len(coords[0]) == 2
+
+    if interiors:
+        assert len(coords) == 3
+    else:
+        assert len(coords) == 1
+        assert len(coords[0]) == 1
 
 
 def test_polygons_with_holes():
@@ -683,38 +815,34 @@ def test_polygons_with_holes():
     assert len(polys) == 4
 
     lens = [len(p) for p in polys]
-    assert lens == [1, 1, 2, 3]
+    assert lens == [3, 2, 1, 1]
 
-    # Pin some values from polygon 3
-    outer, inner1, inner2 = polys[3]
-    assert len(outer) == 21
-    assert len(inner1) == 9  # octagon
-    assert len(inner2) == 5  # diamond
+    # Pin some values from polygon 0
+    outer, inner1, inner2 = polys[0]
+    assert len(outer) == 12
+    assert len(inner1) == 5  # octagon
+    assert len(inner2) == 9  # diamond
 
 
 @pytest.mark.parametrize("iso_value", [0, 0.001, 0.5, 0.9, 1])
 def test_polygon_iso_value(island_slice: MetDataArray, iso_value: float) -> None:
     geojson = island_slice.to_polygon_feature(iso_value=iso_value, epsilon=0.1)
     coords = geojson["geometry"]["coordinates"]
+    assert len(coords[0][0]) == 9
 
-    if iso_value == 1:
-        assert coords == []
-    elif iso_value == 0.9:
-        assert len(coords[0][0]) == 5
-    else:
-        assert len(coords[0][0]) == 9
+
+@pytest.fixture()
+def median_binary_slice(median_binary: MetDataArray) -> MetDataArray:
+    return MetDataArray(median_binary.data.isel(time=[0]))
 
 
 @pytest.mark.skipif(not OPEN3D_AVAILABLE, reason="Open3D not available")
 @pytest.mark.parametrize("return_type", ["geojson", "mesh"])
-def test_polyhedra_save(median_binary: MetDataArray, return_type: str) -> None:
-    assert len(median_binary.data["time"]) == 1
+def test_polyhedra_save(median_binary_slice: MetDataArray, return_type: str) -> None:
+    assert len(median_binary_slice.data["time"]) == 1
     path: str | pathlib.Path
-    if return_type == "geojson":
-        path = "test_mesh.json"
-    else:
-        path = "test_mesh.ply"
-    median_binary.to_polyhedra(return_type=return_type, path=path)
+    path = "test_mesh.json" if return_type == "geojson" else "test_mesh.ply"
+    median_binary_slice.to_polyhedra(return_type=return_type, path=path)
 
     path = pathlib.Path(path)
     assert path.is_file()
@@ -782,6 +910,21 @@ def test_polyhedra_geojson(sparse_binary: tuple[MetDataArray, dict]) -> None:
     assert geojson_altitude.issubset(mda_altitude)
 
 
+@pytest.mark.skipif(not OPEN3D_AVAILABLE, reason="Open3D not available")
+def test_polyhedra_lower_upper_bound(sparse_binary: tuple[MetDataArray, dict]) -> None:
+    """Test the `to_polyhedra` method with lower and upper bounds on interior values."""
+
+    mda, _ = sparse_binary
+    mda1 = MetDataArray(mda.data.isel(time=[0]))
+    mda2 = MetDataArray(-mda.data.isel(time=[0]))
+
+    geojson1 = mda1.to_polyhedra(iso_value=0.5, lower_bound=True)
+    geojson2 = mda2.to_polyhedra(iso_value=-0.5, lower_bound=True)
+    geojson3 = mda2.to_polyhedra(iso_value=-0.5, lower_bound=False)
+    assert geojson1 != geojson2
+    assert geojson1 == geojson3
+
+
 def test_save_load(met_ecmwf_pl_path: str, met_era5_fake: MetDataset) -> None:
     """Test the `save` and `load` methods."""
     _cache = DiskCacheStore(cache_dir=f"{DISK_CACHE_DIR}/test", allow_clear=True)
@@ -797,7 +940,8 @@ def test_save_load(met_ecmwf_pl_path: str, met_era5_fake: MetDataset) -> None:
     assert _cache.exists(f"{mds.hash}-0.nc")
 
     mds2 = MetDataset.load(mds.hash, cachestore=_cache)
-    assert "new" in mds2 and np.all(mds2.data["new"].values == 1)
+    assert "new" in mds2
+    assert np.all(mds2.data["new"].values == 1)
 
     # met data array version
     mda = MetDataArray(mds.data["t"], cachestore=_cache)
@@ -814,7 +958,8 @@ def test_save_load(met_ecmwf_pl_path: str, met_era5_fake: MetDataset) -> None:
     # save multiple time slices separately - dataset
     mds = met_era5_fake.copy()
     mds.cachestore = _cache
-    assert not _cache.exists(f"{mds.hash}-0.nc") and not _cache.exists(f"{mds.hash}-1.nc")
+    assert not _cache.exists(f"{mds.hash}-0.nc")
+    assert not _cache.exists(f"{mds.hash}-1.nc")
     mds.save()
 
     assert _cache.exists(f"{mds.hash}-0.nc")
@@ -829,7 +974,8 @@ def test_save_load(met_ecmwf_pl_path: str, met_era5_fake: MetDataset) -> None:
     mda = MetDataArray(met_era5_fake.data["air_temperature"])
     mda.cachestore = _cache
     assert mda.hash != mds.hash
-    assert not _cache.exists(f"{mda.hash}-0.nc") and not _cache.exists(f"{mda.hash}-1.nc")
+    assert not _cache.exists(f"{mda.hash}-0.nc")
+    assert not _cache.exists(f"{mda.hash}-1.nc")
     mda.save()
 
     assert _cache.exists(f"{mda.hash}-0.nc")
@@ -868,6 +1014,7 @@ def test_met_copy(met_ecmwf_pl_path: str) -> None:
     assert round(mds2["t"].data[5][5][2][0].item()) == 235
 
 
+@pytest.mark.usefixtures("_dask_single_threaded")
 def test_met_wrap_longitude_chunks(met_ecmwf_pl_path: str, override_cache: DiskCacheStore) -> None:
     """Check that the wrap_longitude method increments longitudinal chunks."""
     # Using the xr.open_dataset method
@@ -1013,7 +1160,8 @@ def test_met_variable() -> None:
     NewParam = MetVariable(ecmwf_id=829, standard_name="new-param", short_name="wt", units="n")
     assert isinstance(NewParam, MetVariable)
     assert NewParam.short_name == "wt"
-    assert NewParam.ecmwf_link is not None and "grib/param-db?id=829" in NewParam.ecmwf_link
+    assert NewParam.ecmwf_link is not None
+    assert "grib/param-db?id=829" in NewParam.ecmwf_link
 
     NewParam2 = MetVariable(standard_name="new-param2", short_name="wt", units="n")
     assert NewParam2.ecmwf_link is None
@@ -1028,8 +1176,10 @@ def test_met_variable() -> None:
     # should have attrs property
     NewParam = MetVariable(ecmwf_id=829, standard_name="new-param", short_name="wt", units="n")
     attrs = NewParam.attrs
-    assert "standard_name" in attrs and attrs["standard_name"] == "new-param"
-    assert "short_name" in attrs and attrs["short_name"] == "wt"
+    assert "standard_name" in attrs
+    assert attrs["standard_name"] == "new-param"
+    assert "short_name" in attrs
+    assert attrs["short_name"] == "wt"
     assert "long_name" not in attrs
     assert "ecmwf_id" not in attrs
 
@@ -1064,6 +1214,7 @@ def test_met_ensure_vars(met_ecmwf_pl_path: str) -> None:
     assert variables == ["r", "t"]
 
 
+@pytest.mark.filterwarnings("ignore:Longitude is not evenly spaced")
 def test_include_altitude(met_ecmwf_pl_path: str):
     """Check `to_polygon_feature` with `include_altitude=True`."""
     ds = xr.open_dataset(met_ecmwf_pl_path)
@@ -1078,6 +1229,7 @@ def test_include_altitude(met_ecmwf_pl_path: str):
                 assert coord[2] == 11037.0
 
 
+@pytest.mark.filterwarnings("ignore:Longitude is not evenly spaced")
 def test_include_altitude_raises(met_ecmwf_sl_path: str):
     """Check `to_polygon_feature` raises a ValueError when `include_altitude=True` and SL data."""
     ds = xr.open_dataset(met_ecmwf_sl_path)
@@ -1093,7 +1245,8 @@ def test_include_altitude_raises(met_ecmwf_sl_path: str):
                 assert len(coord) == 2
 
 
-def test_to_polygon_feature_collection(met_ecmwf_pl_path: str):
+@pytest.mark.filterwarnings("ignore:Longitude is not evenly spaced")
+def test_to_polygon_feature_collection(met_ecmwf_pl_path: str) -> None:
     """Confirm that `to_polygon_feature_collection` works."""
     ds = xr.open_dataset(met_ecmwf_pl_path)
     mds = MetDataset(ds[dict(time=[0])])
@@ -1108,10 +1261,24 @@ def test_to_polygon_feature_collection(met_ecmwf_pl_path: str):
         assert feature["properties"]
 
 
+@pytest.mark.filterwarnings("ignore:Longitude is not evenly spaced")
+def test_to_polygon_feature_collection_lower_upper_bound(met_ecmwf_pl_path: str) -> None:
+    """Test `to_polygon_feature_collection` with lower vs upper bound."""
+    ds = xr.open_dataset(met_ecmwf_pl_path)
+    mds = MetDataset(ds[dict(time=[0])])
+    mda1 = mds["q"]
+    mda2 = MetDataArray(-mds["q"].data)
+
+    fc1 = mda1.to_polygon_feature_collection(iso_value=0.0001, lower_bound=True)
+    fc2 = mda2.to_polygon_feature_collection(iso_value=-0.0001, lower_bound=False)
+    assert json.dumps(fc1, sort_keys=True) == json.dumps(fc2, sort_keys=True)
+
+
 def test_wrap_longitude_new_patterns():
     """Test new wrap_longitude patterns (pycontails 0.26.0)."""
+    rng = np.random.default_rng()
     da1 = xr.DataArray(
-        data=np.random.random((360, 5, 5, 1)),
+        data=rng.random((360, 5, 5, 1)),
         dims=["longitude", "latitude", "level", "time"],
         coords={
             "longitude": np.arange(-180, 180),
@@ -1173,3 +1340,66 @@ def test_originates_from_ecmwf_sparse_binary(sparse_binary: tuple[MetDataArray,
     ds = da.to_dataset()
     mds = MetDataset(ds)
     assert not originates_from_ecmwf(mds)
+
+
+@pytest.mark.parametrize("provider", ["ECMWF", "NCEP"])
+def test_provider_attr(met_ecmwf_sl_path: str, provider: str) -> None:
+    """Test the `provider_attr` property with a known provider."""
+
+    ds = xr.open_dataset(met_ecmwf_sl_path)
+    mds = MetDataset(ds, provider=provider)
+    assert mds.provider_attr == provider
+
+
+@pytest.mark.parametrize("provider", ["ECWMF", "NOAA"])
+def test_provider_attr_warning(met_ecmwf_sl_path: str, provider: str) -> None:
+    """Test the `provider_attr` property with an unknown provider."""
+
+    ds = xr.open_dataset(met_ecmwf_sl_path)
+    mds = MetDataset(ds)
+
+    with pytest.raises(KeyError, match=r"Specify 'provider' attribute on underlying dataset."):
+        _ = mds.provider_attr
+
+    mds.attrs["provider"] = provider
+    with pytest.warns(UserWarning, match=f"Unknown provider '{provider}'."):
+        assert mds.provider_attr == provider
+
+
+@pytest.mark.parametrize("coord", ["air_pressure", "altitude"])
+def test_vertical_coord(met_ecmwf_pl_path: str, coord: str) -> None:
+    """Test the nature of assigning new vertical coordinates in the MetDataset constructor."""
+    ds = xr.open_dataset(met_ecmwf_pl_path)
+    assert coord not in ds
+
+    # The MetDataset constructor adds the vertical coordinate
+    mds = MetDataset(ds)
+    assert coord in mds.data.coords
+
+    # Assign a fictitious vertical coordinate with the wrong dtype
+    # and the MetDataset constructor will fix it
+    fake_coord = xr.DataArray(
+        np.arange(len(ds["level"]), dtype=np.float32),
+        dims=["level"],
+        coords={"level": ds["level"]},
+    )
+    ds = ds.assign_coords({coord: fake_coord})
+    mds = MetDataset(ds)
+    assert np.all(mds.data.coords[coord] == fake_coord)
+    assert mds.data.coords[coord].values.dtype == np.float64
+
+
+@pytest.mark.parametrize("copy", [True, False])
+def test_float32_level(met_ecmwf_pl_path: str, copy: bool) -> None:
+    """Confirm that the level coordinate is cast to float64 or an error is raised."""
+    ds = MetDataset(xr.open_dataset(met_ecmwf_pl_path)).data
+
+    ds["level"] = ds["level"].astype(np.float32)
+
+    if copy:
+        mds = MetDataset(ds, copy=copy)
+        assert mds.data["level"].dtype == np.float64
+        return
+
+    with pytest.raises(ValueError, match=r"Level values must have dtype <class 'numpy.float64'>"):
+        MetDataset(ds, copy=copy)
