@@ -2,6 +2,8 @@
 
 from __future__ import annotations
 
+from unittest.mock import patch
+
 import numpy as np
 import pandas as pd
 import pytest
@@ -33,39 +35,43 @@ def test_ISSR_met_source(met_issr: MetDataset) -> None:
         ISSR(met=met_issr)
 
     model = ISSR(met=met_issr, humidity_scaling=ExponentialBoostHumidityScaling())
-    out = model.eval()
+    out1 = model.eval()
 
-    assert isinstance(out, MetDataArray)
-    assert isinstance(out.data, xr.DataArray)
+    assert isinstance(out1, MetDataset)
+    assert isinstance(out1.data, xr.Dataset)
     assert isinstance(model.met, MetDataset)
-    assert out.name == "issr"
+    assert set(out1.data) == {"air_temperature", "specific_humidity", "rhi", "issr"}
+    assert all(v == np.float32 for v in out1.data.dtypes.values())
+    assert out1["issr"].name == "issr"
 
-    assert out.attrs["humidity_scaling_name"] == "exponential_boost"
-    assert out.attrs["humidity_scaling_formula"] == "rhi -> (rhi / rhi_adj) ^ rhi_boost_exponent"
-    assert out.attrs["humidity_scaling_rhi_adj"] == 0.97
-    assert out.attrs["humidity_scaling_rhi_boost_exponent"] == 1.7
-    assert out.attrs["humidity_scaling_clip_upper"] == 1.7
+    assert out1.attrs["humidity_scaling_name"] == "exponential_boost"
+    assert out1.attrs["humidity_scaling_formula"] == "rhi -> (rhi / rhi_adj) ^ rhi_boost_exponent"
+    assert out1.attrs["humidity_scaling_rhi_adj"] == 0.97
+    assert out1.attrs["humidity_scaling_rhi_boost_exponent"] == 1.7
+    assert out1.attrs["humidity_scaling_clip_upper"] == 1.7
 
-    assert out.attrs["description"] == "Ice super-saturated regions"
-    assert "pycontrails_version" in out.attrs
-    assert "met_source" in out.attrs
+    assert "pycontrails_version" in out1.attrs
+    assert out1.attrs["met_source_provider"] == "ECMWF"
+    assert out1.attrs["met_source_dataset"] == "ERA5"
+    assert out1.attrs["met_source_product"] == "reanalysis"
 
     # Pin some counts
-    vals, counts = np.unique(out.data, return_counts=True)
+    vals, counts = np.unique(out1.data["issr"], return_counts=True)
     np.testing.assert_array_equal(vals, [0, 1])
     np.testing.assert_array_equal(counts, [563, 157])
 
     met_copy = met_issr.copy()
     del met_copy.data["air_temperature"]
-    out = model.eval(source=met_copy)
+    out2 = model.eval(source=met_copy)
 
     # Important: if specific_humidity included in model input, it will NOT
     # undergo humidity scaling. This is why the counts below are lower.
-    assert "humidity_scaling" not in out.attrs
-    assert isinstance(out, MetDataArray)
+    assert "humidity_scaling" not in out2.attrs
+    assert isinstance(out2, MetDataset)
+    assert all(v == np.float32 for v in out2.data.dtypes.values())
 
     # Pin some counts
-    vals, counts = np.unique(out.data, return_counts=True)
+    vals, counts = np.unique(out2.data["issr"], return_counts=True)
     np.testing.assert_array_equal(vals, [0, 1])
     np.testing.assert_array_equal(counts, [587, 133])
 
@@ -79,13 +85,23 @@ def test_SAC_met_source(met_issr: MetDataset) -> None:
     model = SAC(met_issr, humidity_scaling=ConstantHumidityScaling())
     out1 = model.eval()
 
-    assert isinstance(out1, MetDataArray)
-    assert isinstance(out1.data, xr.DataArray)
+    assert isinstance(out1, MetDataset)
+    assert isinstance(out1.data, xr.Dataset)
     assert isinstance(model.met, MetDataset)
-    assert out1.name == "sac"
+    assert all(v == np.float32 for v in out1.data.dtypes.values())
+    assert set(out1.data) == {
+        "G",
+        "T_sat_liquid",
+        "rhi",
+        "rh_critical_sac",
+        "specific_humidity",
+        "rh",
+        "sac",
+        "air_temperature",
+    }
 
     # Pin some counts
-    vals, counts = np.unique(out1.data, return_counts=True)
+    vals, counts = np.unique(out1.data["sac"], return_counts=True)
     np.testing.assert_array_equal(vals, [0, 1])
     np.testing.assert_array_equal(counts, [351, 369])
 
@@ -94,24 +110,25 @@ def test_SAC_met_source(met_issr: MetDataset) -> None:
     met_copy.data["engine_efficiency"] = 0.35
     out2 = model.eval(source=met_copy)
 
-    assert isinstance(out2, MetDataArray)
-    assert out2.name == "sac"
-    assert np.all(model.source.data["engine_efficiency"].values == 0.35)
+    assert isinstance(out2, MetDataset)
+    assert np.all(out2.data["engine_efficiency"] == 0.35)
+    assert all(v == np.float32 for v in out1.data.dtypes.values())
 
     # Pin some counts
     # Values are now different because humidity was not scaled
-    vals, counts = np.unique(out2.data, return_counts=True)
+    vals, counts = np.unique(out2.data["sac"], return_counts=True)
     np.testing.assert_array_equal(vals, [0, 1])
-    np.testing.assert_array_equal(counts, [340, 380])
+    np.testing.assert_array_equal(counts, [338, 382])
 
 
 def test_SAC_with_nan(met_issr: MetDataset) -> None:
     """Check that NaN values persist after SAC calculation."""
-    met = met_issr.copy()
+    met = MetDataset(met_issr.data.copy(deep=True))
     met.data["specific_humidity"][4, 3, 2, 1] = np.nan
 
     model = SAC(met, humidity_scaling=ExponentialBoostHumidityScaling())
-    out = model.eval()
+    out = model.eval()["sac"]
+    assert out.data.dtype == np.float32
     np.testing.assert_array_equal(np.nonzero(np.isnan(out.values)), [[4], [3], [2], [1]])
 
     actual = np.unique(out.data)
@@ -128,43 +145,42 @@ def test_PCR_grid(met_issr: MetDataset) -> None:
     model = PCR(met_issr, humidity_scaling=ConstantHumidityScaling())
     out1 = model.eval()
 
-    assert isinstance(out1, MetDataArray)
-    assert isinstance(out1.data, xr.DataArray)
+    assert isinstance(out1, MetDataset)
+    assert isinstance(out1.data, xr.Dataset)
     assert isinstance(model.met, MetDataset)
-    assert out1.name == "pcr"
+    assert all(v == np.float32 for v in out1.data.dtypes.values())
 
     assert out1.attrs.pop("pycontrails_version")
     assert out1.attrs == {
-        "description": "Persistent contrail regions",
         "humidity_scaling_name": "constant_scale",
         "humidity_scaling_formula": "rhi -> rhi / rhi_adj",
         "humidity_scaling_rhi_adj": 0.97,
-        "met_source": "ERA5",
+        "met_source_provider": "ECMWF",
+        "met_source_dataset": "ERA5",
+        "met_source_product": "reanalysis",
         "engine_efficiency": 0.3,
+        "rhi_adj": 0.97,
     }
 
     # Pin some counts
-    vals, counts = np.unique(out1.data, return_counts=True)
+    vals, counts = np.unique(out1.data["pcr"], return_counts=True)
     np.testing.assert_array_equal(vals, [0, 1])
     np.testing.assert_array_equal(counts, [607, 113])
 
     met_copy = met_issr.copy()
     del met_copy.data["air_temperature"]
-    met_copy.data["engine_efficiency"] = 0.35
+    met_copy.data["engine_efficiency"] = (
+        met_copy.data.dims,
+        np.full(met_copy.shape, 0.35, dtype=np.float32),
+    )
     out2 = model.eval(source=met_copy)
 
-    assert isinstance(out2, MetDataArray)
-    assert out2.name == "pcr"
-    assert np.all(model.source.data["engine_efficiency"].values == 0.35)
-
-    assert out2.attrs.pop("pycontrails_version")
-    assert out2.attrs == {
-        "description": "Persistent contrail regions",
-        "met_source": "ERA5",
-    }
+    assert isinstance(out2, MetDataset)
+    assert all(v == np.float32 for v in out2.data.dtypes.values())
+    assert np.all(out2.data["engine_efficiency"] == 0.35)
 
     # Pin some counts
-    vals, counts = np.unique(out2.data, return_counts=True)
+    vals, counts = np.unique(out2.data["pcr"], return_counts=True)
     np.testing.assert_array_equal(vals, [0, 1])
     np.testing.assert_array_equal(counts, [622, 98])
 
@@ -174,9 +190,10 @@ def test_era5_as_expected(met_era5_fake: MetDataset) -> None:
 
     # created with exactly 20% is ISSR
     assert not originates_from_ecmwf(met_era5_fake)
-    issr = ISSR(met=met_era5_fake).eval()
+    issr = ISSR(met=met_era5_fake).eval()["issr"]
     assert isinstance(issr, MetDataArray)
     assert issr.proportion == 0.2
+    assert issr.data.dtype == np.result_type(*met_era5_fake.data.dtypes.values())
 
     # ISSR exist exactly when longitude is a multiple of 5
     # demonstrating this behavior on a few arbitrary selections
@@ -188,8 +205,9 @@ def test_era5_as_expected(met_era5_fake: MetDataset) -> None:
 
     # ISSR and PCR do not completely agree for this data
     # In particular, on levels 150, 175, and 200, the two diverge
-    pcr = PCR(met_era5_fake).eval()
+    pcr = PCR(met_era5_fake).eval()["pcr"]
     assert isinstance(pcr, MetDataArray)
+    assert pcr.data.dtype == np.result_type(*met_era5_fake.data.dtypes.values())
     pcr_low = pcr.data.isel(level=slice(3, None))
     issr_low = issr.data.isel(level=slice(3, None))
     xr.testing.assert_equal(pcr_low, issr_low)
@@ -213,12 +231,10 @@ def test_ISSR_flight_high_altitude(met_era5_fake: MetDataset, flight_fake: Fligh
     fl = flight_fake.copy()
     fl["altitude"] *= 2.5
 
-    with pytest.warns(UserWarning, match="Flight altitude is high"):
-        with pytest.raises(ValueError, match="One of the requested xi is out of bounds"):
-            ISSR(met=met_era5_fake, interpolation_bounds_error=True).eval(source=fl)
+    with pytest.raises(ValueError, match="One of the requested xi is out of bounds"):
+        ISSR(met=met_era5_fake, interpolation_bounds_error=True).eval(source=fl)
 
-    with pytest.warns(UserWarning, match="Flight altitude is high"):
-        fl = ISSR(met=met_era5_fake).eval(source=fl)
+    fl = ISSR(met=met_era5_fake).eval(source=fl)
 
     # waypoints above highest-altitude level in era5 get filled with NaN
     assert isinstance(fl, Flight)
@@ -227,32 +243,48 @@ def test_ISSR_flight_high_altitude(met_era5_fake: MetDataset, flight_fake: Fligh
     assert ~np.isnan(fl["issr"][~low_level_waypoints]).all()
 
 
-def test_ISSR_flight_outside_met(met_era5_fake: MetDataset, flight_fake: Flight) -> None:
-    """Test ISSR model with flight data outside bounds of met data."""
-    # modifying fixture
+@pytest.mark.parametrize("interpolation_use_indices", [True, False])
+def test_ISSR_flight_outside_met(
+    met_era5_fake: MetDataset, flight_fake: Flight, interpolation_use_indices: bool
+) -> None:
+    """Test ISSR model with flight data outside bounds of met data.
+
+    The flight trajectory leaves the met domain in two places:
+
+    - the initial flight waypoints are at low altitude (below the max pressure level in met)
+    - the later flight waypoints have times beyond the max time in met
+    """
+
+    # Alter the flight time
     fl = flight_fake.copy()
     with pytest.warns(UserWarning, match="Overwriting data in key `time`."):
-        fl["time"] += pd.Timedelta("1H")
+        fl["time"] += pd.Timedelta(hours=1)
 
-    with pytest.raises(ValueError, match="One of the requested xi is out of bounds"):
-        _ = ISSR(met_era5_fake, interpolation_bounds_error=True).eval(source=fl)
+    model = ISSR(met_era5_fake, interpolation_use_indices=interpolation_use_indices)
 
-    fl = ISSR(met_era5_fake).eval(source=fl)
+    with pytest.raises(ValueError, match="One of the requested xi is out of bounds in dimension 2"):
+        model.eval(source=fl, interpolation_bounds_error=True)
 
-    # waypoints above highest-altitude level in era5 get filled with NaN
+    # Evaluate the model
+    fl = model.eval(source=fl, interpolation_bounds_error=False)
     assert isinstance(fl, Flight)
-    waypoints_beyond_met_era5_fake = (
-        fl["time"].astype(int) > met_era5_fake.data["time"].max().item()
-    )
-    assert np.isnan(fl["issr"][waypoints_beyond_met_era5_fake]).all()
+    issr = fl["issr"]
+    assert issr.dtype == np.float64
 
-    # any remaining NaN are caused by low altitude waypoints
-    remaining_nan = np.isnan(fl["issr"][~waypoints_beyond_met_era5_fake])
-    low_level_waypoints = (
-        fl["altitude"][~waypoints_beyond_met_era5_fake]
-        < met_era5_fake.data["altitude"].min().item()
-    )
-    assert np.all(remaining_nan == low_level_waypoints)
+    # Waypoints with timestamp beyond the max time are nan-filled
+    late_waypoints = fl["time"] > met_era5_fake["time"].values[-1]
+    assert np.sum(late_waypoints) == 130
+    assert np.isnan(issr[late_waypoints]).all()
+
+    # Waypoints with low altitude are also nan-filled
+    low_waypoints = fl.altitude < met_era5_fake["altitude"].values.min()
+    assert np.sum(low_waypoints) == 257
+    assert np.isnan(fl["issr"][late_waypoints]).all()
+
+    # Anything remaining is non-nan
+    remaining = ~late_waypoints & ~low_waypoints
+    assert np.sum(remaining) == 113
+    assert np.all(np.isfinite(issr[remaining]))
 
 
 def test_ISSR_flight_span_antimeridian(met_era5_fake: MetDataset, flight_fake: Flight) -> None:
@@ -287,14 +319,14 @@ def test_ISSR_flight_fill_value(met_era5_fake: MetDataset, flight_fake: Flight)
 
     # null result (default fill value)
     out = ISSR(met_era5_fake).eval(source=flight_fake)
-    assert np.isnan(out["issr"][low_altitude_waypoints]).all()
-    assert np.isfinite(out["issr"][~low_altitude_waypoints]).all()
+    assert np.all(np.isnan(out["issr"][low_altitude_waypoints]))
+    assert np.all(np.isfinite(out["issr"][~low_altitude_waypoints]))
 
     # When we pass in a custom fill_value here, we no longer encounter nan
     # keep the interpolation fill value reasonable here; else some of our
     # thermo functions will encounter divide by zero issues
     out2 = ISSR(met_era5_fake, interpolation_fill_value=180).eval(source=flight_fake)
-    assert np.isfinite(out2["issr"]).all()
+    assert np.all(np.isfinite(out2["issr"]))
 
 
 def test_ISSR_flight_interp(met_era5_fake: MetDataset, flight_fake: Flight) -> None:
@@ -302,11 +334,11 @@ def test_ISSR_flight_interp(met_era5_fake: MetDataset, flight_fake: Flight) -> N
 
     # test fake method
     with pytest.raises(ValueError, match="Method 'fake' is not defined"):
-        _ = ISSR(met_era5_fake, interpolation_method="fake").eval(source=flight_fake)
+        ISSR(met_era5_fake, interpolation_method="fake").eval(source=flight_fake)
 
     # test bounds error
     with pytest.raises(ValueError, match="xi is out of bounds"):
-        _ = ISSR(met_era5_fake, interpolation_bounds_error=True).eval(source=flight_fake)
+        ISSR(met_era5_fake, interpolation_bounds_error=True).eval(source=flight_fake)
 
     # default interpolation method is linear
     out2 = ISSR(met_era5_fake).eval(source=flight_fake)
@@ -366,8 +398,7 @@ def test_ISSR_flight_missing_waypoint(met_era5_fake: MetDataset, flight_fake: Fl
     fl["latitude"][212] = np.nan
     fl["time"][217] = np.datetime64("NaT")
 
-    with pytest.warns(UserWarning, match="NaT times"):
-        out2 = ISSR(met_era5_fake).eval(source=fl)
+    out2 = ISSR(met_era5_fake).eval(source=fl)
 
     # expect issr column to have exactly at waypoints with NaN
     for idx in [200, 210, 212, 217]:
@@ -444,14 +475,14 @@ def test_PCR_flight(met_era5_fake: MetDataset, flight_fake: Flight) -> None:
     assert np.nansum(out2["pcr"]) == 71
 
 
-@pytest.mark.parametrize("step", [1e-3, 1e-6])
+@pytest.mark.parametrize("step", [1e-3, 1e-4, 1e-5, 1e-6])
 def test_e_sat_liquid_prime(step: float) -> None:
-    """Confirm `sac._e_sat_liquid_prime` formula agrees with secant approximation.
+    """Confirm `thermo.e_sat_liquid_prime` formula agrees with secant approximation.
 
-    This test as well as the `sac._e_sat_liquid_prime` function will need to be
+    This test as well as the `thermo.e_sat_liquid_prime` function will need to be
     updated if `thermo.e_sat_liquid` changes.
     """
-    T0 = np.linspace(-60, -20, 200) - constants.absolute_zero
+    T0 = np.linspace(-60.0, -20.0, 200) - constants.absolute_zero
     assert np.all(T0 > 0)
 
     T1 = T0 + step * np.ones_like(T0)
@@ -460,10 +491,10 @@ def test_e_sat_liquid_prime(step: float) -> None:
     e_sat1 = thermo.e_sat_liquid(T1)
 
     secant_slope = (e_sat1 - e_sat0) / (T1 - T0)
-    tangent_slope = sac._e_sat_liquid_prime(T0)
+    tangent_slope = thermo.e_sat_liquid_prime(T0)
 
     error = np.abs(secant_slope - tangent_slope)
-    assert np.all(error < step)  # approximation is O(step)
+    assert np.all(error < 2 * step)  # approximation is << step
 
 
 def test_T_critical_sac_implementation() -> None:
@@ -510,26 +541,32 @@ def test_T_sat_liquid() -> None:
 
 def test_T_critical_sac_schumann() -> None:
     """Confirm Schumann's statements from his 1996 paper."""
-    rng = np.random.default_rng(737)
-    size = 10000
-    G = rng.uniform(low=1, high=4, size=size)
 
-    # Intentionally going above 1 for rh here; these supersaturated values
-    # get clipped in sac.T_critical_sac
-    rh = rng.uniform(low=0, high=1.01, size=size)
-    t_LM = sac.T_sat_liquid(G)
-    t_LC = sac.T_critical_sac(t_LM, rh, G)
+    with (
+        patch("pycontrails.physics.thermo.e_sat_liquid", thermo.sonntag_e_sat_liquid),
+        patch("pycontrails.physics.thermo.e_sat_liquid_prime", thermo.sonntag_e_sat_liquid_prime),
+    ):
+        rng = np.random.default_rng(737)
+        size = 10000
+        G = rng.uniform(low=1, high=4, size=size)
 
-    # Critical temperature is always below tangent line temperature value
-    assert np.all(t_LC <= t_LM)
-    # Equality when U >= 1
-    idx = np.nonzero(rh >= 1)
-    np.testing.assert_array_equal(t_LC[idx], t_LM[idx])
+        # Intentionally going above 1 for rh here; these supersaturated values
+        # get clipped in sac.T_critical_sac
+        rh = rng.uniform(low=0, high=1.01, size=size)
+        t_LM = sac.T_sat_liquid(G)
+        t_LC = sac.T_critical_sac(t_LM, rh, G)
 
-    # For U = 0, equation (11) gives explicit answer
-    t_LC_U0 = sac.T_critical_sac(t_LM, np.zeros_like(rh), G)
-    eqn11 = t_LM - thermo.e_sat_liquid(t_LM) / G
-    np.testing.assert_array_equal(t_LC_U0, eqn11)
+        # Critical temperature is always below tangent line temperature value
+        assert np.all(t_LC <= t_LM)
+        # Equality when U >= 1
+        idx = np.nonzero(rh >= 1)
+        np.testing.assert_array_equal(t_LC[idx], t_LM[idx])
+
+        # For U = 0, equation (11) gives explicit answer
+        t_LC_U0 = sac.T_critical_sac(t_LM, np.zeros_like(rh), G)
+
+        eqn11 = t_LM - thermo.sonntag_e_sat_liquid(t_LM) / G
+        np.testing.assert_array_equal(t_LC_U0, eqn11)
 
 
 @pytest.mark.parametrize("in_bounds", [False, True])
@@ -552,7 +589,30 @@ def test_issr_source_grid_different_resolution(met_issr: MetDataset, in_bounds:
     if in_bounds:
         out = model.eval(source, interpolation_bounds_error=True)
         assert np.all(np.isfinite(out.data))
+        return
 
+    with pytest.raises(ValueError, match="221"):
+        model.eval(source, interpolation_bounds_error=True)
+
+
+@pytest.mark.filterwarnings(r"ignore:\sMet data appears to have originated from ECMWF")
+@pytest.mark.parametrize("q_method", [None, "cubic-spline"])
+def test_ISSR_grid_variable_q_method(met_issr: MetDataset, q_method: None | str) -> None:
+    """Test ISSR model with variable q_method."""
+
+    source = MetDataset.from_coords(
+        longitude=met_issr.coords["longitude"],
+        latitude=met_issr.coords["latitude"],
+        level=[240, 250, 260],  # different from met_issr
+        time=met_issr.coords["time"],
+    )
+
+    model = ISSR(met_issr, interpolation_q_method=q_method)
+    out = model.eval(source)["issr"]
+
+    assert isinstance(out, MetDataArray)
+    if q_method is None:
+        assert out.data.mean().item() == pytest.approx(0.16805, abs=1e-5)
     else:
-        with pytest.raises(ValueError, match="A value in x_new is below the interpolation range"):
-            model.eval(source, interpolation_bounds_error=True)
+        assert q_method == "cubic-spline"
+        assert out.data.mean().item() == pytest.approx(0.15555, abs=1e-5)
